"""
Source file for readmanganato.com
"""
# pylint: disable=invalid-name

import re

import requests
from bs4 import BeautifulSoup

from .base_source import BaseSource, Chapter, MangaMetadata


class MangaNatoSource(BaseSource):
    name = "MangaNato"
    domains = ["https://manganato.com", "https://readmanganato.com"]
    headers = {"Referer": "https://readmanganato.com/"}

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)
        self._scripts: str | None = None

    def fetch_metadata(self) -> MangaMetadata:
        soup = BeautifulSoup(self._get_scripts(), "html.parser")
        title: str = soup.h1.next_element
        authors_genres = soup.select("td > a.a-h")
        authors: list[str] = []
        genres: list[str] = []
        for el in authors_genres:
            if el["href"].startswith("https://manganato.com/author"):
                authors.append(el.next_element)
            elif el["href"].startswith("https://manganato.com/genre"):
                genres.append(el.next_element)

        cover_img = next(iter(soup.select("span.info-image > img.img-loading")))
        cover_art: str = cover_img["src"]

        description = soup.select_one(
            "#panel-story-info-description > h3"
        ).nextSibling.text.strip()

        return MangaMetadata(title, authors, self.url, genres, description, cover_art)

    def fetch_chapter_list(self) -> list[Chapter]:
        soup = BeautifulSoup(self._get_scripts(), "html.parser")
        chapters = list(
            map(
                lambda c: Chapter(self, c.next_element, c["href"], self.headers),
                soup.select("a.chapter-name"),
            )
        )
        chapters.reverse()
        return chapters

    def fetch_chapter_image_list(self, chapter: Chapter) -> list[str]:
        soup = BeautifulSoup(requests.get(chapter.url).text, "html.parser")
        images = []
        for i in soup.find_all("img"):
            if not i["src"].startswith("https://readmanganato.com"):
                images.append(i["src"])
        return images

    def _get_scripts(self) -> str:
        if self._scripts:
            return self._scripts

        self._scripts = requests.get(self.url).text
        return self._scripts

    @classmethod
    def url_to_id(cls, url: str) -> str:
        *_, last_item = filter(None, url.split("/"))
        return last_item

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(re.match(r"https://readmanganato.com/.*", url))


def get_class() -> type[BaseSource]:
    return MangaNatoSource



#!/usr/bin/env python3

import importlib.metadata
import os
from enum import Enum
from pathlib import Path
from typing import Callable, Iterable, List, Optional

import requests
import typer

import mandown
from mandown.converter import Converter
from mandown.processing import ProcessOps
from mandown.sources.base_source import BaseSource, Chapter, MangaMetadata

app = typer.Typer()


class ConvertFormats(str, Enum):
    CBZ = "cbz"
    EPUB = "epub"
    MOBI = "mobi"
    PDF = "pdf"
    FOLDER = "folder"


# helper functions to repeat less code
def cli_query(url: str) -> BaseSource | None:
    typer.echo(f"Searching sources for {url}")
    try:
        source: BaseSource = mandown.query(url, populate=True)
        typer.secho(f'Found item from source "{source.name}"', fg=typer.colors.GREEN)
        typer.secho(source)
    except ValueError as err:
        typer.secho("Could not match URL with available sources.", fg=typer.colors.RED)
        raise typer.Exit(1) from err

    return source


def cli_convert(
    folder_path: str,
    target_format: ConvertFormats,
    dest: Path = Path(os.getcwd()),
    metadata: MangaMetadata | None = None,
    chapter_list: list[tuple[str, str]] | None = None,
) -> None:
    converter = Converter(folder_path, metadata, chapter_list)
    convert_func: Callable[[str], Iterable] = lambda i: None
    match target_format:
        case ConvertFormats.EPUB:
            convert_func = converter.to_epub_progress
        case ConvertFormats.CBZ:
            convert_func = converter.to_cbz_progress
        case ConvertFormats.PDF:
            convert_func = converter.to_pdf_progress
        case _:
            raise NotImplementedError(
                f"{target_format} conversion has not been implemented yet!"
            )
    with typer.progressbar(
        convert_func(dest),
        length=converter.max_operations[target_format.value],
        label="Converting",
    ) as progress:
        for _ in progress:
            pass

    dest_file = (dest / Path(folder_path).stem).with_suffix(f".{target_format.value}")
    typer.secho(f"Successfully converted to {dest_file}", fg=typer.colors.GREEN)


def cli_process(
    folder_paths: list[Path], options: list[ProcessOps], maxthreads: int = 4
) -> None:
    # only goes down one folder level atm
    if ProcessOps.NO_POSTPROCESSING not in options:
        total_files = sum(len(os.listdir(path)) for path in folder_paths)
        typer.secho(
            f"Applying processing options: {', '.join(options)} to {total_files} images"
        )

        # TODO: move async/multiprocessing into iohandler or processing
        # if processing probably rename iohandler to downloader
        with typer.progressbar(
            mandown.process_progress(folder_paths, options, maxthreads),
            length=total_files,
            label="Processing",
        ) as progress:
            for _ in progress:
                pass


@app.command()
def process(
    options: list[ProcessOps],
    folder_path: Path = typer.Argument(Path(os.getcwd())),
    maxthreads: int = typer.Option(
        4,
        "--threads",
        "-t",
        help="The maximum number of images to process in parallel",
    ),
) -> None:
    """
    Process a folder of images recursively in-place.
    """
    image_paths = list(map(Path, filter(os.path.isdir, folder_path.iterdir())))
    cli_process(image_paths, options, maxthreads)


@app.command()
def convert(
    convert_to: ConvertFormats,
    folder_path: Path,
    metadata_source: Optional[str] = typer.Option(
        None, "--metadata-from", "-m", help="The source to embed metadata from"
    ),
    dest: Path = typer.Option(
        os.getcwd(), "--dest", "-d", help="The folder to save the converted file to."
    ),
    maxthreads: int = typer.Option(
        4,
        "--threads",
        "-t",
        help="The maximum number of images to process in parallel",
    ),
) -> None:
    """
    Convert a folder of images into a comic a la KCC. Optionally
    embed metadata from an online source.
    """
    metadata = cli_query(metadata_source).metadata if metadata_source else None
    cli_convert(folder_path, convert_to, dest, metadata)


@app.callback(invoke_without_command=True, no_args_is_help=True)
def callback(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        is_eager=True,
        help="Display the current version of mandown",
    ),
    supported_sites: Optional[bool] = typer.Option(
        None,
        "--supported-sites",
        is_eager=True,
        help="Output a list of domains supported by mandown",
    ),
) -> None:
    if version:
        typer.echo(f"mandown {importlib.metadata.version('mandown')}")
        raise typer.Exit()

    if supported_sites:
        for source in mandown.sources.get_all_classes():
            typer.echo(f"{source.name}: {', '.join(source.domains)}")
        raise typer.Exit()


# pylint: disable=unused-argument
@app.command()
def download(
    url: str,
    dest: str = typer.Argument(
        os.getcwd(), help="The destination folder to download to."
    ),
    convert_to: ConvertFormats = typer.Option(
        "folder", "--convert", "-c", help="The format to download the comic as"
    ),
    start: Optional[int] = typer.Option(
        None,
        "--start",
        "-s",
        help="The first chapter to download [default: first found]",
    ),
    end: Optional[int] = typer.Option(
        None, "--end", "-e", help="The last chapter to download [default: last found]"
    ),
    maxthreads: int = typer.Option(
        4,
        "--threads",
        "-t",
        help="The maximum number of images to download in parallel",
    ),
    processing_options: Optional[List[ProcessOps]] = typer.Option(
        [],
        "--processing-options",
        "-p",
        help="Image processing options (in-place)",
        case_sensitive=True,
    ),
) -> None:
    """
    Download from a URL chapters start_chapter to end_chapter.
    Defaults to the first chapter and last chapter, respectively
    in the working directory.
    """
    if not os.path.isdir(dest):
        raise ValueError(f"{dest} is not a valid folder path.")

    # get metadata
    source = cli_query(url)

    # starting to think that immutability is much better than whatever
    # the heck is going on here
    target_path = os.path.join(dest, source.metadata.title)
    if not os.path.isdir(target_path):
        os.mkdir(target_path)

    # get processing range
    start_chapter = start or 1
    end_chapter = end or len(source.chapters)

    # zero-index
    start_chapter -= 1

    # download
    chapter_range: list[Chapter] = []
    # yikes these should be split into their own functions sometime
    # get cover art
    with open(Path(target_path) / "cover.jpg", "w+b") as file:
        file.write(requests.get(source.metadata.cover_art).content)

    chapter_range = source.chapters[start_chapter:end_chapter]
    typer.echo("Downloading...")
    for i, chapter in enumerate(chapter_range):
        with typer.progressbar(
            mandown.download_chapter_progress(chapter, target_path, maxthreads),
            length=len(chapter.images),
            label=f"{chapter.title} ({i+1}/{len(chapter_range)})",
        ) as progress:
            for _ in progress:
                pass
    typer.secho(
        f"Successfully downloaded {len(chapter_range)} chapters.",
        fg=typer.colors.GREEN,
    )

    # process
    if processing_options:
        image_paths = list(map(Path, filter(os.path.isdir, os.listdir(target_path))))
        cli_process(image_paths, processing_options)

    # convert
    if convert_to != ConvertFormats.FOLDER:
        cli_convert(
            target_path,
            convert_to,
            dest,
            source.metadata,
            [(c.title, c.title_sanitised) for c in chapter_range],
        )


def main() -> None:
    app()


if __name__ == "__main__":
    main()



import textwrap
from dataclasses import InitVar, dataclass, field
from typing import Callable, Final
import re


class SourceNotOverriddenError(Exception):
    pass


@dataclass(frozen=True)
class MangaMetadata:
    title: str
    authors: list[str]
    url: str
    genres: list[str]
    description: str
    cover_art: str


@dataclass
class Chapter:
    # quotes are used to prevent syntax error
    # pylint: disable=used-before-assignment
    source: InitVar["BaseSource"]  # type: ignore
    title: str
    url: str
    headers: dict[str, str] | None = None

    _image_fetcher: Callable[
        ["Chapter"], list[str]  # type: ignore # pylint: disable=undefined-variable
    ] = field(init=False)
    _images: list[str] = field(init=False, default_factory=list)

    def __post_init__(self, source: "BaseSource") -> None:  # type: ignore
        self._image_fetcher = source.fetch_chapter_image_list
        # get rid of all forbidden chars in filenames
        self.title_sanitised = re.sub(r"[\\\/:*?<>|]", "_", self.title)

    @property
    def images(self) -> list[str]:
        if self._images:
            return self._images
        self._images = self._image_fetcher(self)
        return self._images

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return textwrap.dedent(
            f"""\
        title: {self.title}
        id/url: {self.url}
        images: {self._images}
        headers: {self.headers}\
        """
        )


class BaseSource:
    name = "Source name goes here"
    domains = ["Source domains goes here"]

    _metadata: MangaMetadata | None = None
    _chapters: list[Chapter] = []
    USER_AGENT: Final = (
        "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:77.0) Gecko/20100101 Firefox/77.0"
    )

    def __init__(self, url: str):
        """
        An object representing a manga from a specific source.
        Properties `metadata` and `chapters` will be lazily populated on first access.
        """
        self.url = url

    @property
    def metadata(self) -> MangaMetadata:
        if self._metadata:
            return self._metadata
        self._metadata = self.fetch_metadata()
        return self._metadata

    @property
    def chapters(self) -> list[Chapter]:
        if self._chapters:
            return self._chapters
        self._chapters = self.fetch_chapter_list()
        return self._chapters

    def fetch_metadata(self) -> MangaMetadata:
        """
        Fetch and return title, author, and url of the manga.
        """
        raise SourceNotOverriddenError("Metadata getter not overridden")

    def fetch_chapter_list(self) -> list[Chapter]:
        """
        Fetch and return a list of chapter titles and their URL in ascending order.
        """
        raise SourceNotOverriddenError("Chapter list getter not overridden")

    def fetch_chapter_image_list(self, chapter: Chapter) -> list[str]:
        """
        Given a chapter link, fetch and return a list of image URLs in ascending
        order for that chapter.
        """
        raise SourceNotOverriddenError("Image URL getter not overridden")

    @staticmethod
    def check_url(url: str) -> bool:
        """
        Returns whether the url given matches that of the site and is processable
        """
        raise SourceNotOverriddenError("URL checker not overridden")

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return textwrap.dedent(
            f"""\
            title: {self.metadata.title}
            authors: {self.metadata.authors}
            url: {self.metadata.url}
            genres: {self.metadata.genres}
            description:
                {(chr(10) + " "*16).join(self.metadata.description.split(chr(10)))}
            chapters: {len(self.chapters)}\
            """
        )


def get_class() -> type[BaseSource]:
    return BaseSource



"""
Source file for mangakakalot.com
"""
# pylint: disable=invalid-name

import re

import requests
from bs4 import BeautifulSoup

from .base_source import BaseSource, Chapter, MangaMetadata


class MangaNatoSource(BaseSource):
    name = "MangaKakalot"
    domains = ["https://mangakakalot.com"]
    headers = {"Referer": "https://mangakakalot.com/"}

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)
        self._scripts: str | None = None

    def fetch_metadata(self) -> MangaMetadata:
        soup = BeautifulSoup(self._get_scripts(), "html.parser")
        title: str = soup.h1.next_element
        authors_genres = soup.select(".manga-info-text > li > a")
        authors: list[str] = []
        genres: list[str] = []
        for el in authors_genres:
            if el["href"].startswith("https://mangakakalot.com/search/author"):
                authors.append(el.next_element)
            elif el["href"].startswith("https://mangakakalot.com/manga_list"):
                genres.append(el.next_element)

        cover_img = next(iter(soup.select(".manga-info-pic > img")))
        cover_art: str = cover_img["src"]

        description_html = soup.select_one("#noidungm")
        description_html.find("p").replace_with("")  # remove p block
        description = description_html.get_text().strip()

        return MangaMetadata(title, authors, self.url, genres, description, cover_art)

    def fetch_chapter_list(self) -> list[Chapter]:
        soup = BeautifulSoup(self._get_scripts(), "html.parser")
        chapters = list(
            map(
                lambda c: Chapter(self, c.next_element, c["href"], self.headers),
                soup.select(".row > span > a"),
            )
        )
        chapters.reverse()
        return chapters

    def fetch_chapter_image_list(self, chapter: Chapter) -> list[str]:
        soup = BeautifulSoup(requests.get(chapter.url).text)
        images = []
        for i in soup.select(".container-chapter-reader > img"):
            images.append(i["src"])
        return images

    def _get_scripts(self) -> str:
        if self._scripts:
            return self._scripts

        self._scripts = requests.get(self.url).text
        return self._scripts

    @classmethod
    def url_to_id(cls, url: str) -> str:
        *_, last_item = filter(None, url.split("/"))
        return last_item

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(re.match(r"https://mangakakalot.com/manga/.*", url))


def get_class() -> type[BaseSource]:
    return MangaNatoSource



"""
Source file for webtoons.com
"""
# pylint: disable=invalid-name

import re

import feedparser
import requests
from bs4 import BeautifulSoup

from .base_source import BaseSource, Chapter, MangaMetadata


class WebtoonsSource(BaseSource):
    name = "Webtoons"
    domains = ["https://webtoons.com"]
    headers = {"Referer": "https://webtoons.com/"}

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self._soup: BeautifulSoup | None = None

        title_no_index = url.index("?title_no=") + len("?title_no=")
        title_no_end_index: int | None = url.find("&", title_no_index)
        if title_no_end_index == -1:
            title_no_end_index = None

        self._title_no = int(url[title_no_index:title_no_end_index])

        self._title_path = "/".join(url.split("/")[3:6])

    def fetch_metadata(self) -> MangaMetadata:
        feed = feedparser.parse(
            f"https://www.webtoons.com/{self._title_path}/rss?title_no={self._title_no}"
        )
        feed = feedparser.parse(
            f"https://www.webtoons.com/{self._title_path}/rss?title_no={self._title_no}"
        )
        authors: list[str] = feed["entries"][0].author.split("/")
        authors = list(map(lambda s: s.strip(), authors))
        title: str = feed["channel"]["title"]
        cover_art: str = feed["channel"]["image"]["href"]
        description: str = feed["channel"]["description"].strip()

        return MangaMetadata(
            title,
            authors,
            f"https://www.webtoons.com/{self._title_path}/list?title_no={self._title_no}",
            ["Webtoons.com does not support genres"],
            description,
            cover_art,
        )

    def fetch_chapter_list(self) -> list[Chapter]:
        soup = self._get_soup()
        titles = [
            str(e.next_element) for e in soup.select("p.sub_title > span.ellipsis")
        ]

        links: list[str] = [e["href"] for e in soup.select('a[class^="NPI=a:list"]')]

        chapters = [
            Chapter(self, t, u, {"Referer": "https://webtoons.com"})
            for t, u in zip(titles, links)
        ]
        chapters.reverse()
        return chapters

    def fetch_chapter_image_list(self, chapter: Chapter) -> list[str]:
        soup = BeautifulSoup(requests.get(chapter.url).text, "html.parser")
        images: list[str] = []
        for c in soup.select("div#_imageList > img"):
            images.append(c["data-url"])
        return images

    def _get_soup(self) -> BeautifulSoup:
        if self._soup:
            return self._soup

        # mobile serves all chapters in one page
        mobile_url = (
            f"https://m.webtoons.com/{self._title_path}/list?title_no={self._title_no}"
        )

        self._soup = BeautifulSoup(
            requests.get(
                mobile_url, headers={"Referer": "https://m.webtoons.com"}
            ).text,
            "html.parser",
        )
        return self._soup

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(
            re.match(r"https://www.webtoons.com/.*/list\?title_no=.*", url)
            or re.match(r"https://m.webtoons.com/.*/list\?title_no=.*", url)
            or re.match(r"https://www.webtoons.com/.*/viewer\?title_no=.*", url)
        )


def get_class() -> type[BaseSource]:
    return WebtoonsSource



from enum import Enum
from pathlib import Path
from typing import Iterable

try:
    from PIL import Image, ImageChops

    HAS_PILLOW = True
except ImportError:
    HAS_PILLOW = False


class ProcessOps(str, Enum):
    """
    Constants for Processor.process
    """

    ROTATE_DOUBLE_PAGES = "rotate_double_pages"
    """If page width is greater than its height, rotate it 90 degrees."""
    SPLIT_DOUBLE_PAGES = "split_double_pages"
    """If page width is greater than its height, split it in half into two images."""
    NO_POSTPROCESSING = "none"
    """Disable image processing entirely."""
    TRIM_BORDERS = "trim_borders"
    """Conservatively remove borders in images."""


class Processor:
    def __init__(self, image_path: Path | str, right_to_left: bool = False) -> None:
        if not HAS_PILLOW:
            raise ImportError(
                "Pillow was not found and is needed for processing. Is it installed?"
            )
        self.image_path = Path(image_path)
        self._image = Image.open(self.image_path)
        self.modified = False
        self.right_to_left = right_to_left

        # what if we need to process pending images in future operations?
        self.pending_images: list[tuple[Image.Image, Path]] = []

    @property
    def image(self) -> Image.Image:
        return self._image

    @image.setter
    def image(self, image: Image.Image) -> None:
        self._image = image
        self.modified = True

    def write(self, filename: Path | str | None = None) -> None:
        """Save the processed image manually"""
        self.image.save(filename or self.image_path)

        for image, path in self.pending_images:
            image.save(path)

    def process(
        self, operations: list[ProcessOps], filename: Path | str | None = None
    ) -> None:
        """
        Perform the operations in `operations` on the image in sequence
        and save it to disk. If `filename` is not None, it will be saved
        with that filename instead.
        """
        if ProcessOps.NO_POSTPROCESSING in operations:
            return

        for func in operations:
            try:
                getattr(self, func)()
            except AttributeError as err:
                raise NotImplementedError(
                    f"{func} is not a valid post-processing function."
                ) from err
            except OSError as err:
                raise OSError(f"Error in {self.image_path}") from err

        if self.modified:
            # only write to disk if something has actually changed
            self.write(filename)

    def rotate_double_pages(self) -> None:
        width, height = self.image.size
        if width > height:
            self.image = self.image.rotate(90, expand=1)

    def split_double_pages(self) -> None:
        width, height = self.image.size
        if not width > height:
            return

        left = self.image.crop((0, 0, int(width / 2), height))
        right = self.image.crop((int(width / 2), 0, width, height))

        self.image = right if self.right_to_left else left
        new_image = left if self.right_to_left else right

        new_image_path = self.image_path.with_stem(self.image_path.stem + ".5")
        self.pending_images.append((new_image, new_image_path))

    def trim_borders(self) -> None:
        bg = Image.new(self.image.mode, self.image.size, self.image.getpixel((0, 0)))
        diff = ImageChops.difference(self.image, bg)
        diff = ImageChops.add(diff, diff, 2.0, -100)
        bbox = diff.getbbox()
        if bbox:
            self.image = self.image.crop(bbox)


def async_process(data: tuple[Path | str, list[ProcessOps]]) -> None:
    image_path, ops = data
    processor = Processor(image_path)
    processor.process(ops)



"""
Handles downloading files
"""
# pylint: disable=invalid-name
import imghdr
import multiprocessing as mp
import os
import time
import urllib.parse
from pathlib import Path
from typing import Iterable, Sequence

import requests
import undetected_chromedriver as uc


def async_download(data: tuple[str, str, str | None, dict[str, str] | None]) -> None:
    url, dest_folder, filename, headers = data
    name = filename or url.split("/")[-1]
    dest_file = os.path.join(dest_folder, name)

    response = requests.get(url, headers=headers)
    response.raise_for_status()
    with open(dest_file, "wb") as file:
        file.write(response.content)

    # if the file extension is lying
    # rename it so epubcheck doesn't yell at us
    ext = imghdr.what(dest_file)
    if ext is not None:
        os.rename(dest_file, Path(dest_file).with_suffix(f".{ext}"))


def download(
    urls: Sequence[str] | str,
    dest_folder: str,
    headers: dict[str, str] = None,
    maxthreads: int = 1,
    filestems: Sequence[str] | str | None = None,
) -> Iterable[None]:
    """
    Download one or multiple URLs to a destination folder.
    Raises ValueError if the folder does not exist.
    """
    if isinstance(urls, str):
        urls = [urls]

    if isinstance(filestems, str):
        filestems = [filestems]

    if not os.path.isdir(dest_folder):
        raise ValueError(f"Folder path {dest_folder} does not exist")

    map_pool: list[tuple[str, str, str, dict[str, str] | None]] = []
    padding = f"0{len(str(len(urls)))}"
    if filestems is None:
        filestems = [f"{i+1:{padding}}" for i in range(len(urls))]

    for u, stem in zip(urls, filestems, strict=True):
        _, ext = os.path.splitext(urllib.parse.urlparse(u).path)

        map_pool.append((u, dest_folder, f"{stem}{ext}", headers))

    with mp.Pool(maxthreads) as pool:
        yield from pool.imap_unordered(async_download, map_pool)


class UndetectedDriver:
    def __init__(self) -> None:
        self.driver = uc.Chrome()
        self.first_time = True

    def get(self, url: str) -> str:
        self.driver.get(url)
        if self.first_time:
            time.sleep(5)
            self.first_time = False

        return self.driver.find_element_by_tag_name("html").get_attribute("outerHTML")



"""
Source file for readcomiconline.li
"""
# pylint: disable=invalid-name

import re

import requests
from bs4 import BeautifulSoup
from mandown.iohandler import UndetectedDriver

from .base_source import BaseSource, Chapter, MangaMetadata


class ReadComicOnlineSource(BaseSource):
    name = "ReadComicOnline"
    domains = ["https://readcomiconline.li"]

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)
        self.driver = UndetectedDriver()
        self.title = ""

    def fetch_metadata(self) -> MangaMetadata:
        soup = BeautifulSoup(
            self.driver.get(f"https://readcomiconline.li/Comic/{self.id}"),
            features="lxml",
        )

        title = str(soup.select_one("a.bigChar").text)
        self.title = title
        author = [
            str(
                # this site uses "Various" if there's more than one author
                soup.select_one("a[href^='/Writer']").text
            )
        ]
        genres: list[str] = [str(e.text) for e in soup.select("a[href^='/Genre']")]
        desc_element = soup.select_one("p[style='text-align: justify;']")

        description = str(desc_element.text) if desc_element else ""
        cover = self.domains[0] + str(soup.find("link")["href"])

        return MangaMetadata(title, author, self.url, genres, description, cover)

    def fetch_chapter_list(self) -> list[Chapter]:
        soup = BeautifulSoup(
            self.driver.get(f"https://readcomiconline.li/Comic/{self.id}"),
            features="lxml",
        )

        chapters: list[Chapter] = []
        for e in soup.select("td > a"):
            chapters.append(
                Chapter(
                    self,
                    next(e.children).text.strip()[len(self.title) + 1 :],
                    self.domains[0] + e["href"],
                )
            )
        return list(reversed(chapters))

    def fetch_chapter_image_list(self, chapter: Chapter) -> list[str]:
        text = self.driver.get(chapter.url)

        images: list[str] = []
        start = 0
        while (index := text.find("lstImages.push(", start)) != -1:
            s_index = index + len('lstImages.push("')
            e_index = text.find('");', s_index)
            images.append(text[s_index:e_index])
            start = e_index
        return images

    @classmethod
    def url_to_id(cls, url: str) -> str:
        *_, last_item = filter(None, url.split("/"))
        return last_item

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(re.match(r"https://readcomiconline.li/Comic/.*", url))


def get_class() -> type[BaseSource]:
    return ReadComicOnlineSource



# pylint: disable=invalid-name
import multiprocessing as mp
import os
from pathlib import Path
from typing import Iterable

from natsort import natsorted

from mandown import converter, iohandler, processing, sources
from mandown.processing import ProcessOps
from mandown.sources.base_source import BaseSource, Chapter


def query(url: str, populate: bool = True, populate_sort: bool = True) -> BaseSource:
    """
    Return the source file for a URL.
    """
    source = sources.get_class_for(url)(url)
    if populate:
        # these statements are to trigger their getters to
        # fetch the data
        if source.metadata:
            pass
        if source.chapters:
            if populate_sort:
                titles = list(map(lambda c: c.title, source.chapters))
                if titles != natsorted(titles):
                    padding = f"0{len(str(len(source.chapters)))}"
                    for i, c in enumerate(source.chapters):
                        c.title = f"{i+1:{padding}}. {c.title}"
    return source


def download_chapter_progress(
    chapter: Chapter, dest_folder: str, maxthreads: int = 1, only_needed: bool = True
) -> Iterable[None]:
    """
    Download the images of a chapter to a destination folder.
    Returns a generator that increments whenever an item has finished
    downloading.
    Raises ValueError if the folder does not exist.
    """
    if not chapter.images:
        raise ValueError("No images available to download")

    if not os.path.isdir(dest_folder):
        raise ValueError(f"Folder path {dest_folder} does not exist")

    download_folder = os.path.join(dest_folder, chapter.title)
    if not os.path.isdir(download_folder):
        os.mkdir(download_folder)

    padding = len(str(len(chapter.images)))
    skip_images: set[int] = set()
    if only_needed:
        for f in os.listdir(download_folder):
            name = Path(f).stem
            if name == name.rjust(padding, "0"):
                try:
                    skip_images.add(int(name))
                except ValueError:
                    # expected if it's not an image
                    pass

    if len(skip_images) != len(chapter.images):
        # zip will crash if fed an empty array
        processed_chapter_images, filestems = zip(
            *(
                (link, str(i).rjust(padding, "0"))
                for i, link in enumerate(chapter.images, start=1)
                if i not in skip_images
            )
        )

        yield from iohandler.download(
            processed_chapter_images,
            os.path.join(dest_folder, download_folder),
            chapter.headers,
            maxthreads,
            filestems,
        )


def download_chapter(
    chapter: Chapter, dest_folder: str, maxthreads: int = 1, only_needed: bool = True
) -> None:
    """
    Download the images of a chapter to a destination folder.
    Raises ValueError if the folder does not exist.
    """
    for _ in download_chapter_progress(chapter, dest_folder, maxthreads, only_needed):
        pass


def process_progress(
    folder_paths: list[Path], options: list[ProcessOps], maxthreads: int = 4
) -> Iterable[None]:
    map_pool: list[tuple[Path | str, list[ProcessOps]]] = []
    for folder in folder_paths:
        for image_path in folder.iterdir():
            if image_path.is_file():
                map_pool.append((image_path.absolute(), options))

    with mp.Pool(maxthreads) as pool:
        yield from pool.imap_unordered(processing.async_process, map_pool)


def process(
    folder_paths: list[Path], options: list[ProcessOps], maxthreads: int = 4
) -> None:
    for _ in process_progress(folder_paths, options, maxthreads):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mandown', 'mandown.sources']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.10.0,<5.0.0',
 'feedparser>=6.0.8,<7.0.0',
 'lxml>=4.7.1,<5.0.0',
 'natsort>=8.0.2,<9.0.0',
 'requests>=2.27.0,<3.0.0',
 'typer>=0.4.0,<0.5.0',
 'undetected-chromedriver>=3.1.5,<4.0.0']

extras_require = \
{'postprocessing': ['Pillow>=9.0.1,<10.0.0']}

entry_points = \
{'console_scripts': ['mandown = mandown.cli:main']}

setup_kwargs = {
    'name': 'mandown',
    'version': '0.8.1a1.dev1',
    'description': 'Command line application and library to download and convert comics from various sources',
    'long_description': '# mandown\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n[![Download from PyPI](https://img.shields.io/pypi/v/mandown)](https://pypi.org/project/mandown)\n[![Download from the AUR](https://img.shields.io/aur/version/mandown-git)](https://aur.archlinux.org/packages/mandown-git)\n[![Latest release](https://img.shields.io/github/v/release/potatoeggy/mandown?display_name=tag)](https://github.com/potatoeggy/mandown/releases/latest)\n[![License](https://img.shields.io/github/license/potatoeggy/mandown)](/LICENSE)\n\nComic downloader and converter to CBZ, EPUB, and/or PDF as a Python library and command line application.\n\n## Supported sites\n\nTo request a new site, please file a [new issue](https://github.com/potatoeggy/mandown/issues/new).\n\n- https://mangasee123.com\n- https://manganato.com\n- https://webtoons.com\n- https://mangadex.org\n- https://mangakakalot.com\n- https://readcomiconline.li\n\n## Installation\n\nInstall the package from PyPI:\n\n```\npip3 install mandown\n```\n\nOr, to build from source:\n\nMandown depends on [poetry](https://github.com/python-poetry/poetry) for building.\n\n```\ngit clone https://github.com/potatoeggy/mandown.git\npoetry install\npoetry build\npip3 install dist/mandown*.whl\n```\n\n## Usage\n\n```\nmandown download <URL>\n```\n\nTo convert the download contents to CBZ/EPUB, append the `--convert` option. To apply image processing to the downloaded images, append the `--process` option.\n\n```\nmandown download <URL> --convert epub --process rotate_double_pages\n```\n\nTo convert an existing folder without downloading anything except metadata (like a stripped-down version of https://github.com/ciromattia/kcc), use the `convert` command.\n\n```\nmandown convert <FORMAT> <PATH_TO_FOLDER>\n```\n\nRun `mandown --help` for more info.\n\n## Library usage\n\n```python\nimport os\nimport mandown\nfrom mandown.converter import Converter\n\ncomic = mandown.query(url_to_comic)\nprint(comic.metadata, comic.chapters)\nfor c in comic.chapters:\n    mandown.download_chapter(c, dest_folder=os.getcwd(), maxthreads=4)\n\nfolder_path = os.getcwd()\nmandown.process(folder_path, [ProcessOps.TRIM_BORDERS, ProcessOps.ROTATE_DOUBLE_PAGES])\n\nconverter = Converter(folder_path, metadata=comic.metadata)\nconverter.to_epub()\n```\n',
    'author': 'Daniel Chen',
    'author_email': 'danielchen04@hotmail.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/potatoeggy/mandown',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
Converts a folder of images to EPUB/PDF/CBZ/MOBI
"""
import datetime
import os
import re
import shutil
import tempfile
import textwrap
import unicodedata
import zipfile
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Iterator

from natsort import natsorted

try:
    from PIL import Image

    HAS_PILLOW = True
except ImportError:
    HAS_PILLOW = False

from mandown.sources.base_source import MangaMetadata

PDF_IMAGE_MAX_INTERVAL = 200  # adjust for memory as necessary
PDF_IMAGE_MIN_INTERVAL_FACTOR = 0.12

ACCEPTED_IMAGE_EXTENSIONS = {
    ".gif": "gif",
    ".png": "png",
    ".jpg": "jpeg",
    ".jpeg": "jpeg",
    ".jpe": "jpeg",
}

# temporary measure
KOBO_WIDTH, KOBO_HEIGHT = 1005, 1430


class PageProgression(str, Enum):
    LEFT_TO_RIGHT = "ltr"
    RIGHT_TO_LEFT = "rtl"


@dataclass
class Options:
    page_progression: PageProgression = PageProgression.LEFT_TO_RIGHT


class Converter:
    def __init__(
        self,
        folder_path: str,
        metadata: MangaMetadata | None = None,
        chapter_list: list[tuple[str, str]] | None = None,
        options: Options | None = None,
    ) -> None:
        """
        Get the file path of a comic and identify its metadata.
        The chapter list is a list of tuples in the format
        (`chapter_path`, `chapter_title`), where `chapter_path`
        is the path of the chapter relative to `folder_path`.

        If the chapter list or metadata are not provided, they will
        be guessed from the directory structure.
        """
        if not Path(folder_path).is_dir():
            raise ValueError(f"{folder_path} is not a directory.")

        self.folder_path = Path(os.path.realpath(folder_path))
        self.metadata = metadata or MangaMetadata(
            title=os.path.basename(self.folder_path),
            authors=[],
            url="",
            genres=[],
            description="Generated by mandown",
            cover_art="",
        )

        # self.chapters is found either by traversing the path (natsort) or by being fed
        working_chapters = chapter_list or []
        if not working_chapters:
            for item in natsorted(os.listdir(self.folder_path)):
                item = str(item)  # get rid of type complaints
                if (Path(self.folder_path) / item).is_dir():
                    working_chapters.append((item, item))

        padding = f"0{len(str(len(working_chapters)))}"

        self.chapters: list[tuple[str, str, str, list[str]]] = [
            (
                title,
                sanitised,
                f"{i+1:{padding}}-{self.slugify(sanitised)}",
                list(
                    filter(
                        lambda s: any(
                            str(s).lower().endswith(i)
                            for i in ACCEPTED_IMAGE_EXTENSIONS
                        ),
                        natsorted(os.listdir(self.folder_path / sanitised)),  # type: ignore
                    )
                ),
            )
            for i, (title, sanitised) in enumerate(working_chapters)
        ]

        for file in Path(folder_path).iterdir():
            if file.stem == "cover" and file.suffix in ACCEPTED_IMAGE_EXTENSIONS:
                self.cover = file
                break
        else:
            self.cover = None

        num_files = len(list(Path.rglob(self.folder_path, "*")))
        # useful for progress bars
        self.max_operations: dict[str, int] = {
            "epub": len(self.chapters) * 3,
            "cbz": len(self.chapters),
            "pdf": round(
                num_files
                / min(
                    PDF_IMAGE_MAX_INTERVAL,
                    num_files * PDF_IMAGE_MIN_INTERVAL_FACTOR,
                )
            ),
        }
        self.options = options or Options()

    def to_cbz_progress(self, dest_folder: str) -> Iterator:
        dest_file = Path(dest_folder) / f"{self.metadata.title}.cbz"
        with zipfile.ZipFile(dest_file, "w", zipfile.ZIP_DEFLATED) as file:  # type: ignore
            for dirpath, _, filenames in os.walk(self.folder_path):  # from kcc
                yield "Compressing"
                for name in filenames:
                    if (Path(dirpath) / name).is_file():
                        file.write(
                            Path(dirpath) / name,  # type: ignore
                            Path(dirpath.lstrip(str(self.folder_path))) / name,
                            zipfile.ZIP_DEFLATED,
                        )

    def to_cbz(self, dest_folder: str) -> None:
        for _ in self.to_cbz_progress(dest_folder):
            pass

    def to_epub_progress(self, dest_folder: str) -> Iterator:
        if len(self.chapters) == 0:
            raise IOError(
                f"{self.folder_path} must contain at least one folder for EPUB conversion"
                " for chapter recognition. If there is only one chapter, move your"
                f" images to a new folder in {self.folder_path} and rerun the command."
            )
        # taken from KCC generated epub file structure
        with tempfile.TemporaryDirectory() as temp:
            root = Path(temp)
            EpubGenerator.create_skeleton(root)

            oebps = root / "OEBPS"
            with open(oebps / "Text" / "style.css", "w", encoding="utf-8") as file:
                file.write(EpubGenerator.style_css)

            with open(oebps / "content.opf", "w", encoding="utf-8") as file:
                file.write(
                    EpubGenerator.generate_content_opf(
                        self.metadata,
                        self.chapters,
                        PageProgression.LEFT_TO_RIGHT,
                        self.cover,
                    )
                )

            with open(oebps / "toc.ncx", "w", encoding="utf-8") as file:
                file.write(EpubGenerator.generate_toc_ncx(self.metadata, self.chapters))

            with open(oebps / "nav.xhtml", "w", encoding="utf-8") as file:
                file.write(
                    EpubGenerator.generate_nav_xhtml(self.metadata, self.chapters)
                )

            for _, path, slug, images in self.chapters:
                yield "Building"
                os.mkdir(oebps / "Text" / slug)
                os.mkdir(oebps / "Images" / slug)
                padding = f"0{len(str(len(images)))}"

                for index, i in enumerate(images, start=1):
                    with open(
                        oebps / "Text" / slug / f"{index:{padding}}.xhtml",
                        "w",
                        encoding="utf-8",
                    ) as file:
                        file.write(
                            EpubGenerator.generate_image_html(
                                slug, f"{index:{padding}}{Path(i).suffix}"
                            )
                        )
                    shutil.copyfile(
                        self.folder_path / path / i,
                        oebps / "Images" / slug / f"{index:{padding}}{Path(i).suffix}",
                    )

            if self.cover is not None:
                shutil.copyfile(self.cover, oebps / "Images" / self.cover.name)

            # compress and move epub
            dest_file = Path(dest_folder) / f"{self.metadata.title}.epub"

            with zipfile.ZipFile(dest_file, "w", zipfile.ZIP_DEFLATED) as zipout:
                zipout.writestr("mimetype", "application/epub+zip", zipfile.ZIP_STORED)
                for dirpath, _, filenames in os.walk(root):  # from kcc
                    yield "Compressing"
                    for name in filenames:
                        if (Path(dirpath) / name).is_file():
                            zipout.write(
                                Path(dirpath) / name,
                                Path(dirpath.lstrip(str(root))) / name,
                                zipfile.ZIP_DEFLATED,
                            )

    def to_epub(self, dest_folder: str) -> None:
        for _ in self.to_epub_progress(dest_folder):
            pass

    def to_pdf_progress(self, dest_folder: str) -> Iterator:
        if not HAS_PILLOW:
            raise ImportError(
                "Pillow could not be found and is required for PDF conversion."
            )

        images: list[Image.Image] = []
        for f in Path.rglob(self.folder_path, "*"):
            if f.suffix in ACCEPTED_IMAGE_EXTENSIONS:
                images.append(Image.open(f))

        if len(images) > 0:
            interval = round(
                min(
                    PDF_IMAGE_MAX_INTERVAL,
                    len(images) * PDF_IMAGE_MIN_INTERVAL_FACTOR,
                )
            )

            for i in range(0, len(images), interval):
                images[i].save(
                    Path(dest_folder) / (self.metadata.title + ".pdf"),
                    "PDF",
                    resolution=100.0,
                    save_all=True,
                    append_images=images[i + 1 : i + interval]
                    if len(images) > i + 1
                    else None,
                    title=self.metadata.title,
                    author=""
                    if len(self.metadata.authors) == 0
                    else self.metadata.authors[0],
                    append=i != 0,
                    creator="mandown",
                    producer="mandown",
                )

                for j in images[i : i + interval]:
                    j.close()
                yield

    def to_pdf(self, dest_folder: str) -> None:
        for _ in self.to_pdf_progress(dest_folder):
            pass

    def to_mobi(self) -> None:
        pass

    @staticmethod
    def slugify(string: str) -> str:
        """
        Remove all potentially invalid characters for XML
        """
        slug = unicodedata.normalize("NFKD", string)
        slug = slug.lower()
        slug = re.sub(r"[^a-z0-9]+", "-", slug).strip("-")
        slug = re.sub(r"[-]+", "-", slug)
        return slug


# pylint: disable=line-too-long
class EpubGenerator:
    style_css = textwrap.dedent(
        """\
        @page {
        margin: 0;
        }
        body {
        display: block;
        margin: 0;
        padding: 0;
        }\
        """
    )

    @staticmethod
    def create_skeleton(root: Path) -> None:
        os.mkdir(root / "META-INF")
        with open(root / "META-INF" / "container.xml", "w", encoding="utf-8") as file:
            file.write(
                textwrap.dedent(
                    """\
                    <?xml version="1.0"?>
                    <container version="1.0" xmlns="urn:oasis:names:tc:opendocument:xmlns:container">
                    <rootfiles>
                    <rootfile full-path="OEBPS/content.opf" media-type="application/oebps-package+xml"/>
                    </rootfiles>
                    </container>\
                    """
                )
            )

        os.mkdir(root / "OEBPS")
        os.mkdir(root / "OEBPS" / "Images")
        os.mkdir(root / "OEBPS" / "Text")

    @staticmethod
    def generate_image_html(
        chapter_slug: str,
        image_path: str,
        width: int = KOBO_WIDTH,
        height: int = KOBO_HEIGHT,
    ) -> str:
        image_base = Path(image_path).with_suffix("")
        return textwrap.dedent(
            f"""\
            <?xml version="1.0" encoding="UTF-8"?>
            <!DOCTYPE html>
            <html xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops">
            <head>
            <title>{image_base}</title>
            <link href="../style.css" type="text/css" rel="stylesheet"/>
            </head>
            <body style="">
            <div style="text-align:center;top:0.0%;">
            <img width="{width}" height="{height}" src="../../Images/{chapter_slug}/{image_path}"/>
            </div>
            </body>
            </html>\
            """
        )

    @staticmethod
    def generate_toc_ncx(
        metadata: MangaMetadata, chapters: list[tuple[str, str, str, list[str]]]
    ) -> str:
        nav_map: list[str] = []
        for title, _, slug, images in chapters:
            padding = f"0{(len(str(len(images))))}"
            nav_map.append(
                f'<navPoint id="Text_{slug}"><navLabel><text>{title}</text></navLabel><content src="Text/{slug}/{1:{padding}}.xhtml"/></navPoint>'
            )

        return textwrap.dedent(
            f"""\
            <?xml version="1.0" encoding="UTF-8"?>
            <ncx version="2005-1" xml:lang="en-US" xmlns="http://www.daisy.org/z3986/2005/ncx/">
            <head>
            <meta name="dtb:uid" content="urn:uuid:bafec70e-0286-49d3-b1e2-9e9f297c8cfe"/>
            <meta name="dtb:depth" content="1"/>
            <meta name="dtb:totalPageCount" content="0"/>
            <meta name="dtb:maxPageNumber" content="0"/>
            <meta name="generated" content="true"/>
            </head>
            <docTitle><text>{metadata.title}</text></docTitle>
            <navMap>
            {(chr(10) + " "*12).join(nav_map)}
            </navMap>
            </ncx>\
            """
        )

    @staticmethod
    def generate_content_opf(
        metadata: MangaMetadata,
        chapters: list[tuple[str, str, str, list[str]]],
        progression: PageProgression,
        cover: Path | None = None,
    ) -> str:
        """
        `chapters` must be a list containing a tuple of elements in this order:
          - `title`, the title of the chapter
          - `sanitised_title`, the relative path to the chapter folder
          - `slug`, a safe-for-html version of the title
          - `images`, a list of paths to each image
        """
        # fmt: off
        cover_html = (
            f'<item id="cover" href="Images/{cover.name}" media-type="image/{ACCEPTED_IMAGE_EXTENSIONS[cover.suffix]}" properties="cover-image"/>'
            if cover else ""
        )
        # fmt: on

        time_now = datetime.datetime.utcnow().replace(microsecond=0).isoformat()

        item_ids: list[str] = []
        item_refs: list[str] = []
        for *_, slug, images in chapters:
            padding = f"0{len(str(len(images)))}"
            for index, i in enumerate(images, start=1):
                marker = f"{index:{padding}}"
                item_ids.append(
                    f'<item id="page_Images_{slug}_{marker}" href="Text/{slug}/{marker}.xhtml" media-type="application/xhtml+xml" />'
                )
                item_ids.append(
                    f'<item id="img_Images_{slug}_{marker}" href="Images/{slug}/{marker}{Path(i).suffix}" media-type="image/{ACCEPTED_IMAGE_EXTENSIONS[Path(i).suffix]}" />'
                )

                item_refs.append(f'<itemref idref="page_Images_{slug}_{marker}" />')

        return textwrap.dedent(
            f"""\
            <?xml version="1.0" encoding="UTF-8"?>
            <package version="3.0" unique-identifier="BookID" xmlns="http://www.idpf.org/2007/opf">
            <metadata xmlns:opf="http://www.idpf.org/2007/opf" xmlns:dc="http://purl.org/dc/elements/1.1/">
            <dc:title>{metadata.title}</dc:title>
            <dc:language>en-US</dc:language>
            <dc:identifier id="BookID">urn:uuid:bafec70e-0286-49d3-b1e2-9e9f297c8cfe</dc:identifier>
            <dc:contributor id="contributor">mandown</dc:contributor>
            {(chr(10) + " "*12).join(
                f'<dc:creator id="id-{i}">{name}</dc:creator>'
                for i, name in enumerate(metadata.authors)
            )}
            <dc:description>{
                (chr(10) + " "*12).join(
                    metadata.description.split(chr(10))
                )
            }</dc:description>
            {
                (chr(10) + " "*12).join(
                    f"<dc:subject>{genre}</dc:subject>"
                    for genre in metadata.genres
                )
            }
            <meta property="dcterms:modified">{time_now}Z</meta>
            <meta name="cover" content="cover"/>
            <meta property="rendition:orientation">portrait</meta>
            <meta property="rendition:spread">both</meta>
            <meta property="rendition:layout">pre-paginated</meta>
            </metadata>
            <manifest>
            <item id="ncx" href="toc.ncx" media-type="application/x-dtbncx+xml"/>
            <item id="nav" href="nav.xhtml" properties="nav" media-type="application/xhtml+xml"/>
            <item id="css" href="Text/style.css" media-type="text/css"/>
            {cover_html}
            {(chr(10) + " "*12).join(item_ids)}
            </manifest>
            <spine page-progression-direction="{progression.value}" toc="ncx">
            {(chr(10) + " "*12).join(item_refs)}
            </spine>
            </package>\
            """
        )

    @staticmethod
    def generate_nav_xhtml(
        metadata: MangaMetadata, chapters: list[tuple[str, str, str, list[str]]]
    ) -> str:
        lis: list[str] = []
        for title, _, slug, images in chapters:
            padding = f"0{len(str(len(images)))}"
            lis.append(
                f'<li><a href="Text/{slug}/{1:{padding}}.xhtml">{title}</a></li>'
            )

        return textwrap.dedent(
            f"""\
            <?xml version="1.0" encoding="utf-8"?>
            <!DOCTYPE html>
            <html xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops">
            <head>
            <title>{metadata.title}</title>
            <meta charset="utf-8"/>
            </head>
            <body>
            <nav xmlns:epub="http://www.idpf.org/2007/ops" epub:type="toc" id="toc">
            <ol>
            {(chr(10) + " "*12).join(lis)}
            </ol>
            </nav>
            <nav epub:type="page-list">
            <ol>
            {(chr(10) + " "*12).join(lis)}
            </ol>
            </nav>
            </body>
            </html>\
            """
        )
